from typing import Any

from precis_i18n.codepointset import CodepointSet

class UnicodeData:
    _halfwidth_chars = ...
    _space_chars = ...

    def __init__(self, ucd: Any = ...) -> None: ...
    @property
    def version(self) -> float: ...
    def category(self, char: str) -> str: ...
    def combining(self, char: str) -> int: ...
    def bidirectional(self, char: str) -> str: ...
    def normalize(self, form: str, value: str) -> str: ...
    def width_map(self, value: str) -> str: ...
    def map_nonascii_space_to_ascii(self, value: str) -> str: ...
    def default_ignorable(self, cp: int) -> bool: ...
    def has_compat(self, cp: int) -> bool: ...
    def control(self, cp: int) -> bool: ...
    def noncharacter(self, cp: int) -> bool: ...
    def old_hangul_jamo(self, cp: int) -> bool: ...
    def greek_script(self, cp: int) -> bool: ...
    def hebrew_script(self, cp: int) -> bool: ...
    def hiragana_katakana_han_script(self, cp: int) -> bool: ...
    def combining_virama(self, cp: int) -> bool: ...
    def arabic_indic(self, cp: int) -> bool: ...
    def extended_arabic_indic(self, cp: int) -> bool: ...
    def valid_jointype(self, value: str, offset: int) -> bool: ...

_DEFAULT_IGNORABLE: CodepointSet
_JOINTYPE_DUAL_JOINING: CodepointSet
_JOINTYPE_RIGHT_JOINING: CodepointSet
_JOINTYPE_LEFT_JOINING: CodepointSet
_JOINTYPE_TRANSPARENT: CodepointSet
_GREEK_SCRIPT: CodepointSet
_HEBREW_SCRIPT: CodepointSet
_HIRAGANA_KATAKANA_HAN: CodepointSet
_OLD_HANGUL_JAMO: CodepointSet



from typing import Optional, Union

from precis_i18n.baseclass import BaseClass
from precis_i18n.unicode import UnicodeData

class Profile:
    def __init__(
        self, base: BaseClass, name: str, casemap: Optional[str] = ...
    ) -> None: ...
    @property
    def base(self) -> BaseClass: ...
    @property
    def name(self) -> str: ...
    def enforce(self, value: Union[bytes, str]) -> str: ...
    def apply_five_rules(self, value: str) -> str: ...
    def width_mapping_rule(self, value: str) -> str: ...
    def additional_mapping_rule(self, value: str) -> str: ...
    def case_mapping_rule(self, value: str) -> str: ...
    def normalization_rule(self, value: str) -> str: ...
    def directionality_rule(self, value: str) -> str: ...
    def idempotence_check(self, value: str) -> str: ...

class Username(Profile):
    def __init__(
        self, ucd: UnicodeData, name: str, casemap: Optional[str] = ...
    ) -> None: ...
    def width_mapping_rule(self, value: str) -> str: ...
    def directionality_rule(self, value: str) -> str: ...

class OpaqueString(Profile):
    def __init__(self, ucd: UnicodeData, name: str) -> None: ...
    def additional_mapping_rule(self, value: str) -> str: ...

class Nickname(Profile):
    def __init__(
        self, ucd: UnicodeData, name: str, casemap: Optional[str] = ...
    ) -> None: ...
    def additional_mapping_rule(self, value: str) -> str: ...
    def normalization_rule(self, value: str) -> str: ...
    def idempotence_check(self, value: str) -> str: ...



from typing import Generator, Tuple

class CodepointSet:
    def __init__(self, table: str) -> None: ...
    def __contains__(self, cp: int) -> bool: ...
    def __len__(self) -> int: ...
    def __eq__(self, rhs: object) -> bool: ...
    def __repr__(self) -> str: ...
    def items(self) -> Generator[Tuple[int, int], None, None]: ...



from typing import NoReturn, Optional, Tuple

from precis_i18n.unicode import UnicodeData

class BaseClass:
    ucd: UnicodeData
    name: str

    def __init__(self, ucd: UnicodeData, name: str = ...) -> None: ...
    def enforce(self, value: str, codec_name: Optional[str] = ...) -> str: ...

class IdentifierClass(BaseClass):
    _allowed: Tuple[str]

class FreeFormClass(BaseClass):
    _allowed: Tuple[str, str]

def raise_error(encoding: str, value: str, offset: int, error: str) -> NoReturn: ...



"""Implements the `get_profile` factory function used to create profiles."""

import precis_i18n.baseclass as _base
import precis_i18n.profile as _profile
import precis_i18n.unicode as _unicode


def _factory(profile, **kwds):
    def _construct(ucd):
        return profile(ucd, **kwds)

    return _construct


_PROFILES = {
    "identifierclass": _factory(_base.IdentifierClass, name="IdentifierClass"),
    "freeformclass": _factory(_base.FreeFormClass, name="FreeFormClass"),
    "usernamecasepreserved": _factory(_profile.Username, name="UsernameCasePreserved"),
    "usernamecasemapped": _factory(
        _profile.Username, name="UsernameCaseMapped", casemap="lower"
    ),
    "usernamecasemapped_casefold": _factory(
        _profile.Username, name="UsernameCaseMapped:CaseFold", casemap="fold"
    ),
    "usernamecasemapped_tolower": _factory(
        _profile.Username, name="UsernameCaseMapped:ToLower", casemap="lower"
    ),
    "opaquestring": _factory(_profile.OpaqueString, name="OpaqueString"),
    "nicknamecasepreserved": _factory(_profile.Nickname, name="NicknameCasePreserved"),
    "nicknamecasemapped": _factory(
        _profile.Nickname, name="NicknameCaseMapped", casemap="lower"
    ),
    "nicknamecasemapped_casefold": _factory(
        _profile.Nickname, name="NicknameCaseMapped:CaseFold", casemap="fold"
    ),
    "nicknamecasemapped_tolower": _factory(
        _profile.Nickname, name="NicknameCaseMapped:ToLower", casemap="lower"
    ),
    # Alias for backward-compatibility with previous version of codec.
    "nickname": _factory(_profile.Nickname, name="Nickname", casemap="lower"),
}


def get_profile(name, *, unicodedata=None):
    """Return the desired PRECIS profile object.

    Choose name from:
        "IdentifierClass"
        "FreeFormClass"
        "UsernameCasePreserved"
        "UsernameCaseMapped"
        "UsernameCaseMapped:CaseFold"
        "UsernameCaseMapped:ToLower"
        "OpaqueString"
        "NicknameCasePreserved"
        "NicknameCaseMapped"
        "NicknameCaseMapped:CaseFold"
        "NicknameCaseMapped:ToLower"
        "Nickname" (alias for "NicknameCaseMapped")

    This function constructs a new profile each time; there is no cache.

    To use an alternative Unicode implementation, pass a module or object that
    implements the unicodedata interface via the unicodedata keyword argument.
    The default is to use the unicodedata module built into the Python runtime.

    Args:
        name (str): name of a PRECIS profile
        unicodedata (module|object): Alternative unicodedata interface

    Returns:
        AbstractProfile: PRECIS profile object.

    Raises:
        KeyError: Profile not found.
    """
    profile = name.lower().replace(":", "_")
    return _PROFILES[profile](_unicode.UnicodeData(unicodedata))



from typing import Dict, Union

from precis_i18n.baseclass import BaseClass
from precis_i18n.profile import Profile
from precis_i18n.unicode import UnicodeData

_PROFILES = Dict[str, Union[BaseClass, Profile]]

def get_profile(name: str, *, unicodedata: UnicodeData = ...) -> Profile: ...



from typing import Dict, Tuple

from precis_i18n.unicode import UnicodeData

PVALID: str
FREE_PVAL: str
DISALLOWED: str
UNASSIGNED: str
CONTEXTJ: str
CONTEXTO: str

def derived_property(cp: int, ucd: UnicodeData) -> Tuple[str, str]: ...
def in_letter_digits(category: str) -> bool: ...
def in_exceptions(cp: int) -> bool: ...
def in_backward_compatible(cp: int) -> bool: ...
def in_join_control(cp: int) -> bool: ...
def in_old_hangul_jamo(cp: int, ucd: UnicodeData) -> bool: ...
def in_unassigned(cp: int, category: str, ucd: UnicodeData) -> bool: ...
def in_ascii7(cp: int) -> bool: ...
def in_controls(cp: int, ucd: UnicodeData) -> bool: ...
def in_precis_ignorable_properties(cp: int, ucd: UnicodeData) -> bool: ...
def in_spaces(category: str) -> bool: ...
def in_symbols(category: str) -> bool: ...
def in_punctuation(category: str) -> bool: ...
def in_has_compat(cp: int, ucd: UnicodeData) -> bool: ...
def in_other_letter_digits(category: str) -> bool: ...
def exceptions(cp: int) -> str: ...
def backward_compatible(cp: int) -> str: ...

_EXCEPTIONS_TABLE: Dict[bytes, str]

_BACKWARD_COMPATIBLE_TABLE = ...



"""Implements the PRECIS (RFC 8264) `derived_property` function."""

PVALID = "PVALID"
FREE_PVAL = "FREE_PVAL"
DISALLOWED = "DISALLOWED"
UNASSIGNED = "UNASSIGNED"
CONTEXTJ = "CONTEXTJ"
CONTEXTO = "CONTEXTO"

# pylint: disable=too-many-return-statements,too-many-branches


def derived_property(cp, ucd):  # noqa: C901
    """Return value of the PRECIS derived property of a code point.

    From section 8 of RFC 8264:

       If .cp. .in. Exceptions Then Exceptions(cp);
       Else If .cp. .in. BackwardCompatible Then BackwardCompatible(cp);
       Else If .cp. .in. Unassigned Then UNASSIGNED;
       Else If .cp. .in. ASCII7 Then PVALID;
       Else If .cp. .in. JoinControl Then CONTEXTJ;
       Else If .cp. .in. OldHangulJamo Then DISALLOWED;
       Else If .cp. .in. PrecisIgnorableProperties Then DISALLOWED;
       Else If .cp. .in. Controls Then DISALLOWED;
       Else If .cp. .in. HasCompat Then ID_DIS or FREE_PVAL;
       Else If .cp. .in. LetterDigits Then PVALID;
       Else If .cp. .in. OtherLetterDigits Then ID_DIS or FREE_PVAL;
       Else If .cp. .in. Spaces Then ID_DIS or FREE_PVAL;
       Else If .cp. .in. Symbols Then ID_DIS or FREE_PVAL;
       Else If .cp. .in. Punctuation Then ID_DIS or FREE_PVAL;
       Else DISALLOWED;

    Args:
        cp (int): Code point.
        ucd (UnicodeData): Unicode character database.

    Returns:
        str: Derived property for `cp`.
    """
    category = ucd.category(chr(cp))

    if in_exceptions(cp):
        return exceptions(cp), "exceptions"
    if in_backward_compatible(cp):
        return backward_compatible(cp), "backward_compatible"  # pragma: no cover
    if in_unassigned(cp, category, ucd):
        return UNASSIGNED, "unassigned"
    if in_ascii7(cp):
        return PVALID, "ascii7"
    if in_join_control(cp):
        return CONTEXTJ, "join_control"
    if in_old_hangul_jamo(cp, ucd):
        return DISALLOWED, "old_hangul_jamo"
    if in_precis_ignorable_properties(cp, ucd):
        return DISALLOWED, "precis_ignorable_properties"
    if in_controls(cp, ucd):
        return DISALLOWED, "controls"
    if in_has_compat(cp, ucd):
        return FREE_PVAL, "has_compat"
    if in_letter_digits(category):
        return PVALID, "letter_digits"
    if in_other_letter_digits(category):
        return FREE_PVAL, "other_letter_digits"
    if in_spaces(category):
        return FREE_PVAL, "spaces"
    if in_symbols(category):
        return FREE_PVAL, "symbols"
    if in_punctuation(category):
        return FREE_PVAL, "punctuation"
    return DISALLOWED, "other"


def in_letter_digits(category):
    """Category for code points informally described as "language characters".

    Args:
        category (str): Unicode general category.

    Returns:
        bool: True if `category` in set.
    """
    return category in {"Ll", "Lu", "Lo", "Nd", "Lm", "Mn", "Mc"}


def in_exceptions(cp):
    """Code points for which the derived property cannot be assigned using
    only the Unicode core property values.

    Args:
        cp (int): Code point.

    Returns:
        bool: True if `cp` is in set.
    """
    return cp in _EXCEPTIONS_TABLE


def in_backward_compatible(cp):
    """Code points whose Unicode property values have changed such that their
    derived property changed.

    Args:
        cp (int): Code point.

    Returns:
        bool: True if `cp` is in set.
    """
    return cp in _BACKWARD_COMPATIBLE_TABLE


def in_join_control(cp):
    """Code points for Join Control characters required under some
    circumstances. (CONTEXTJ)

    Args:
        cp (int): Code point.

    Returns:
        bool: True if `cp` is in set.
    """
    return 0x200C <= cp <= 0x200D


def in_old_hangul_jamo(cp, ucd):
    """Code points for all conjoining Hangul Jamo (Leading Jamo, Vowel Jamo,
    and Trailing Jamo).

    Args:
        cp (int): Code point.
        ucd (UnicodeData): Unicode character database.

    Returns:
        bool: True if `cp` is in set.
    """
    return ucd.old_hangul_jamo(cp)


def in_unassigned(cp, category, ucd):
    """Code points that are not (yet) assigned in Unicode.

    Args:
        cp (int): Code point.
        category (str): Unicode general category for `cp`.
        ucd (UnicodeData): Unicode character database.

    Returns:
        bool: True if `cp` is in set.
    """
    return category == "Cn" and not ucd.noncharacter(cp)


def in_ascii7(cp):
    """Code points for all printable, non-space characters from the 7-bit
    ASCII range.

    Args:
        cp (int): Code point.

    Returns:
        bool: True if `cp` is in set.
    """
    return 0x21 <= cp <= 0x7E


def in_controls(cp, ucd):
    """Code points for all control characters.

    Args:
        cp (int): Code point.
        ucd (UnicodeData): Unicode character database.

    Returns:
        bool: True if `cp` is in set.
    """
    return ucd.control(cp)


def in_precis_ignorable_properties(cp, ucd):
    """Code points that are discouraged from use in PRECIS string classes.

    Args:
        cp (int): Code point.
        ucd (UnicodeData): Unicode character database.

    Returns:
        bool: True if `cp` is in set.
    """
    return ucd.default_ignorable(cp) or ucd.noncharacter(cp)


def in_spaces(category):
    """Category for code points that are space characters.

    Args:
        category (str): Unicode general category.

    Returns:
        bool: True if `category` in set.
    """
    return category in {"Zs"}


def in_symbols(category):
    """Category for code points that are symbols.

    Args:
        category (str): Unicode general category.

    Returns:
        bool: True if `category` in set.
    """
    return category in {"Sm", "Sc", "Sk", "So"}


def in_punctuation(category):
    """Category for code points that are punctuation characters.

    Args:
        category (str): Unicode general category.

    Returns:
        bool: True if `category` in set.
    """
    return category in {"Pc", "Pd", "Ps", "Pe", "Pi", "Pf", "Po"}


def in_has_compat(cp, ucd):
    """Code points that have compatibility equivalents.

    Args:
        cp (int): Code point.
        ucd (UnicodeData): Unicode character database.

    Returns:
        bool: True if `cp` is in set.
    """
    return ucd.has_compat(cp)


def in_other_letter_digits(category):
    """Code points that are letters and digits other than the "traditional"
    letters and digits.

    Args:
        category (str): Unicode general category.

    Returns:
        bool: True if `category` in set.
    """
    return category in {"Lt", "Nl", "No", "Me"}


def exceptions(cp):
    """Return derived property for exception codepoint.

    Args:
        cp (int): Code point.

    Returns:
        str: Derived property for `cp`.
    """
    return _EXCEPTIONS_TABLE[cp]


def backward_compatible(cp):  # pragma: no cover
    """Return derived property for backward-compatible code point.

    Args:
        cp (int): Code point.

    Returns:
        str: Derived property for `cp`.
    """
    return _BACKWARD_COMPATIBLE_TABLE[cp]


_EXCEPTIONS_TABLE = {
    # Source: RFC 5892, Section 2.6, pp. 7-8.
    # PVALID -- Would otherwise have been DISALLOWED
    0x00DF: PVALID,  # LATIN SMALL LETTER SHARP S
    0x03C2: PVALID,  # GREEK SMALL LETTER FINAL SIGMA
    0x06FD: PVALID,  # ARABIC SIGN SINDHI AMPERSAND
    0x06FE: PVALID,  # ARABIC SIGN SINDHI POSTPOSITION MEN
    0x0F0B: PVALID,  # TIBETAN MARK INTERSYLLABIC TSHEG
    0x3007: PVALID,  # IDEOGRAPHIC NUMBER ZERO
    # CONTEXTO -- Would otherwise have been DISALLOWED
    0x00B7: CONTEXTO,  # MIDDLE DOT
    0x0375: CONTEXTO,  # GREEK LOWER NUMERAL SIGN (KERAIA)
    0x05F3: CONTEXTO,  # HEBREW PUNCTUATION GERESH
    0x05F4: CONTEXTO,  # HEBREW PUNCTUATION GERSHAYIM
    0x30FB: CONTEXTO,  # KATAKANA MIDDLE DOT
    # CONTEXTO -- Would otherwise have been PVALID
    0x0660: CONTEXTO,  # ARABIC-INDIC DIGIT ZERO
    0x0661: CONTEXTO,  # ARABIC-INDIC DIGIT ONE
    0x0662: CONTEXTO,  # ARABIC-INDIC DIGIT TWO
    0x0663: CONTEXTO,  # ARABIC-INDIC DIGIT THREE
    0x0664: CONTEXTO,  # ARABIC-INDIC DIGIT FOUR
    0x0665: CONTEXTO,  # ARABIC-INDIC DIGIT FIVE
    0x0666: CONTEXTO,  # ARABIC-INDIC DIGIT SIX
    0x0667: CONTEXTO,  # ARABIC-INDIC DIGIT SEVEN
    0x0668: CONTEXTO,  # ARABIC-INDIC DIGIT EIGHT
    0x0669: CONTEXTO,  # ARABIC-INDIC DIGIT NINE
    0x06F0: CONTEXTO,  # EXTENDED ARABIC-INDIC DIGIT ZERO
    0x06F1: CONTEXTO,  # EXTENDED ARABIC-INDIC DIGIT ONE
    0x06F2: CONTEXTO,  # EXTENDED ARABIC-INDIC DIGIT TWO
    0x06F3: CONTEXTO,  # EXTENDED ARABIC-INDIC DIGIT THREE
    0x06F4: CONTEXTO,  # EXTENDED ARABIC-INDIC DIGIT FOUR
    0x06F5: CONTEXTO,  # EXTENDED ARABIC-INDIC DIGIT FIVE
    0x06F6: CONTEXTO,  # EXTENDED ARABIC-INDIC DIGIT SIX
    0x06F7: CONTEXTO,  # EXTENDED ARABIC-INDIC DIGIT SEVEN
    0x06F8: CONTEXTO,  # EXTENDED ARABIC-INDIC DIGIT EIGHT
    0x06F9: CONTEXTO,  # EXTENDED ARABIC-INDIC DIGIT NINE
    # DISALLOWED -- Would otherwise have been PVALID
    0x0640: DISALLOWED,  # ARABIC TATWEEL
    0x07FA: DISALLOWED,  # NKO LAJANYALAN
    0x302E: DISALLOWED,  # HANGUL SINGLE DOT TONE MARK
    0x302F: DISALLOWED,  # HANGUL DOUBLE DOT TONE MARK
    0x3031: DISALLOWED,  # VERTICAL KANA REPEAT MARK
    0x3032: DISALLOWED,  # VERTICAL KANA REPEAT WITH VOICED SOUND MARK
    0x3033: DISALLOWED,  # VERTICAL KANA REPEAT MARK UPPER HALF
    0x3034: DISALLOWED,  # VERTICAL KANA REPEAT WITH VOICED SOUND MARK UPPER HA
    0x3035: DISALLOWED,  # VERTICAL KANA REPEAT MARK LOWER HALF
    0x303B: DISALLOWED,  # VERTICAL IDEOGRAPHIC ITERATION MARK
}

_BACKWARD_COMPATIBLE_TABLE = {}



"""Implements the CodepointSet class."""

import io
import re
from bisect import bisect_left


class CodepointSet:
    """Concrete class for an immutable set of Unicode code points.

    Inclusive ranges [a, b] are stored as adjacent unicode characters in a
    string. The low end of a range has an even index. The high end is at an
    odd index. Singleton ranges are stored as [c, c].

    To test if a code point is in the set, search for the code point. If it
    matches a char in the string, the code point is in the set. If not, find
    the index of the next largest char. If this index is odd, the code point is
    in the set.

    This class is constructed from a multi-line string containing a sequence of
    codepoints and codepoint ranges.

      HHHH
      HHHH..HHHH

    H is a hexadecimal digit. Comment lines begin with '#'. Blank lines are
    ignored.

    Note: Sets with any non-BMP codepoints will use 32-bits for all codepoints.
    (PEP 393 Flexible String Representation)

    Args:
        table (str): Multi-line string of code point ranges.
    """

    def __init__(self, table):
        self._table = _stringify(_coalesce(_parse(table)))
        assert (len(self._table) % 2) == 0

    def __contains__(self, cp):
        """Check if code point `cp` is in the set.

        Args:
            cp (int): Code point.

        Returns:
            bool: True if `cp` is in the set.
        """
        if not 0 <= cp <= 0x10FFFF:
            return False
        char = chr(cp)
        idx = bisect_left(self._table, char)
        if idx >= len(self._table):
            return False
        return (idx % 2) == 1 or self._table[idx] == char

    def __len__(self):
        """Return size of set.

        Used for debugging only.

        Returns:
            int: Count of total code points.
        """
        count = 0
        for lo, hi in self.items():
            count += hi - lo + 1
        return count

    def __eq__(self, rhs):
        """Check if set is equal to other set.

        Args:
            rhs (CodepointSet): Other set.

        Returns:
            bool: True if sets are equal.
        """
        # pylint: disable=protected-access
        if self.__class__ != rhs.__class__:
            return False
        return self._table == rhs._table

    def __repr__(self):
        """Return string representation of set.

        Example:
            "CodepointSet('0000\n0010..00FF')"
        """
        elems = "\\n".join(_repr(elem) for elem in self.items())
        return "CodepointSet('%s')" % elems

    def items(self):
        """Generator yielding sequence of range tuples (lo, hi)."""
        for i in range(len(self._table) // 2):
            lo = ord(self._table[2 * i])
            hi = ord(self._table[2 * i + 1])
            yield (lo, hi)


def _parse(table):
    """Parse a multi-line string containing codepoint ranges.

    Args:
        table (str): Multi-line string of code point ranges.

    Returns:
        list: List of 2-tuples (lo, hi) representing code point ranges.

    Raises:
        ValueError: Error while parsing `table`.
    """
    codepoint = re.compile(r"^([0-9A-Fa-f]+)(?:\.\.([0-9A-Fa-f]+))?$")
    elems = []
    for line in io.StringIO(table):
        line = line.strip()
        m = codepoint.match(line)
        if not m:
            if line and line[0] != "#":
                raise ValueError("Unable to parse line: %s" % line)
            continue
        lo = int(m.group(1), 16)
        hi = int(m.group(2), 16) if m.group(2) else lo
        if lo > hi:
            raise ValueError("Invalid range (lo > hi): %s" % line)
        elems.append((lo, hi))
    return elems


def _coalesce(elems):
    """Sort and coalesce adjacent ranges (in-place).

    Args:
        elems (list): List of 2-tuples (lo, hi).

    Returns:
        list: List `elems` after sorting in-place and coalescing ranges.

    Raises:
        ValueError: Overlapping ranges.
    """
    elems.sort()
    i = 0
    while i < len(elems) - 1:
        (lo0, hi0), (lo1, hi1) = elems[i : i + 2]
        if not lo0 <= hi0 < lo1 <= hi1:
            raise ValueError("Range overlaps at index %d: %r" % (i, elems[i : i + 2]))
        if lo1 == hi0 + 1:
            elems[i : i + 2] = [(lo0, hi1)]
        else:
            i += 1
    return elems


def _stringify(elems):
    """Convert a sequence of ranges into a string.

    Args:
        elems (list): List of 2-tuples representing ranges.

    Returns:
        str: String with lo..hi ranges concatenated.
    """
    return "".join(chr(lo) + chr(hi) for (lo, hi) in elems)


def _repr(elem):
    """Return string representation for tuple (lo, hi).

    Examples:
        "HHHH"
        "HHHH..HHHH"

    Args:
        elem (tuple): 2-tuple (lo, hi)

    Returns:
        str: String representation of range tuple.
    """
    if elem[0] == elem[1]:
        return "%04X" % elem[0]
    return "%04X..%04X" % elem



"""Implements the BiDi Rule.

(Source: RFC 5893, Section 2)

The following rule, consisting of six conditions, applies to labels
in Bidi domain names.  The requirements that this rule satisfies are
described in Section 3.  All of the conditions must be satisfied for
the rule to be satisfied.

1. The first character must be a character with Bidi property L, R,
   or AL.  If it has the R or AL property, it is an RTL label; if it
   has the L property, it is an LTR label.

2. In an RTL label, only characters with the Bidi properties R, AL,
   AN, EN, ES, CS, ET, ON, BN, or NSM are allowed.

3. In an RTL label, the end of the label must be a character with
   Bidi property R, AL, EN, or AN, followed by zero or more
   characters with Bidi property NSM.

4. In an RTL label, if an EN is present, no AN may be present, and
   vice versa.

5. In an LTR label, only characters with the Bidi properties L, EN,
   ES, CS, ET, ON, BN, or NSM are allowed.

6. In an LTR label, the end of the label must be a character with
   Bidi property L or EN, followed by zero or more characters with
   Bidi property NSM.
"""

_LTR_FIRST = {"L"}
_LTR_ALLOWED = {"L", "EN", "ES", "CS", "ET", "ON", "BN", "NSM"}
_LTR_LAST = {"L", "EN"}
_LTR_EXCL = set()

_RTL_FIRST = {"R", "AL"}
_RTL_ALLOWED = {"R", "AL", "AN", "EN", "ES", "CS", "ET", "ON", "BN", "NSM"}
_RTL_LAST = {"R", "AL", "EN", "AN"}
_RTL_EXCL = {"EN", "AN"}

_RTL_ANY = {"R", "AL", "AN"}


def bidi_rule(value, ucd):
    """Check if `value` obeys the BiDi rule.

    Args:
        value (str): String value to check.
        ucd (UnicodeData): Unicode character database.

    Returns:
        bool: True if value satisfies BiDi rule.
    """
    bidi = ucd.bidirectional(value[0])
    if bidi in _LTR_FIRST:
        return _bidi_rule(value, ucd, _LTR_ALLOWED, _LTR_LAST, _LTR_EXCL)
    if bidi in _RTL_FIRST:
        return _bidi_rule(value, ucd, _RTL_ALLOWED, _RTL_LAST, _RTL_EXCL)
    return False


def _bidi_rule(value, ucd, allowed, last, exclusive):
    """Check the bidi_rule for LTR or RTL, depending on parameters.

    Args:
        value (str): String value to check.
        ucd (UnicodeData): Unicode character database.
        allowed (set): Set of allowed BiDi properties.
        last (set): Set of BiDi properties allowed at end (followed by NSM).
        exclusive (set): Set of BiDi properties that are mutually exclusive.

    Returns:
        bool: True if value satisfies the BiDi rule.
    """
    assert ucd.bidirectional(value[0]) in _LTR_FIRST | _RTL_FIRST

    # Starting from the end, find the first character whose bidi is not 'NSM'.
    bidi = None
    found = -1
    for i in reversed(range(len(value))):
        bidi = ucd.bidirectional(value[i])
        if bidi != "NSM":
            found = i
            break

    # Last non-NSM character must be in `last`.
    if found < 0 or bidi not in last:
        return False

    # Check if last char is in the exclusive set.
    bidi_seen = bidi if bidi in exclusive else None

    # Make sure the remaining characters are allowed.
    for i in range(1, found):
        bidi = ucd.bidirectional(value[i])
        if bidi not in allowed:
            return False
        if bidi in exclusive and bidi_seen != bidi:
            if bidi_seen:
                return False
            bidi_seen = bidi

    return True


def has_rtl(value, ucd):
    """Check if value contains any RTL characters.

    Args:
        value (str): String value to check.
        ucd (UnicodeData): Unicode character database.

    Returns:
        bool: True if value contains RTL characters.
    """
    return any(ucd.bidirectional(x) in _RTL_ANY for x in value)



"""Registers precis_i18n codec."""

import codecs

from precis_i18n import get_profile


def _make_encode(profile):
    def _encode(input, errors="strict"):
        if errors != "strict":
            raise ValueError("invalid errors argument")
        return (profile.enforce(input).encode("utf-8"), len(input))

    return _encode


def _not_supported(input, errors="strict"):
    # pylint: disable=unused-argument
    raise NotImplementedError("decode not supported")


def search(name):
    """Search function registered for PRECIS codecs.

    Args:
        name (str): Codec name.

    Returns:
        CodecInfo: Encode/decode information or None if not found.
    """
    try:
        profile = get_profile(name)
    except KeyError:
        return None

    return codecs.CodecInfo(
        name=name, encode=_make_encode(profile), decode=_not_supported
    )


codecs.register(search)



"""Implements PRECIS rules for derived properties CONTEXTJ and CONTEXTO."""


def context_rule_error(value, offset, ucd):
    """Apply the PRECIS context rules to `value[offset]`.

    Args:
        value (str): String value to check.
        offset (int): Position within `value`.
        ucd (UnicodeData): Unicode character database.

    Returns:
        str: '' if no error, or name of the rule that failed.
    """
    cp = ord(value[offset])
    if ucd.arabic_indic(cp):
        rule = rule_arabic_indic
    elif ucd.extended_arabic_indic(cp):
        rule = rule_extended_arabic_indic
    else:
        rule = _RULES[cp]

    try:
        valid = rule(value, offset, ucd)
    except IndexError:
        # Handle failure of _before and _after accessors.
        valid = False

    if valid:
        return ""

    # If context rule fails, return name of context rule (the name of the
    # function with 'rule_' prefix removed.)
    result = rule.__name__
    if result.startswith("rule_"):
        result = result[5:]
    return result


# These rules test a character at a given offset in the string.


def rule_zero_width_nonjoiner(value, offset, ucd):
    """Return true if context permits a ZERO WIDTH NON-JOINER (U+200C).

    From https://tools.ietf.org/html/rfc5892#appendix-A.1:

      "This may occur in a formally cursive script (such as Arabic) in a
      context where it breaks a cursive connection as required for
      orthographic rules, as in the Persian language, for example.  It
      also may occur in Indic scripts in a consonant-conjunct context
      (immediately following a virama), to control required display of
      such conjuncts."

    Args:
        value (str): String value to check.
        offset (int): Position within `value`.
        ucd (UnicodeData): Unicode character database.

    Returns:
        bool: True if value is allowed.
    """
    assert value[offset] == "\u200c"
    if ucd.combining_virama(_before(value, offset)):
        return True
    if ucd.valid_jointype(value, offset):
        return True
    return False


def rule_zero_width_joiner(value, offset, ucd):
    """Return true if context permits a ZERO WIDTH JOINER (U+200D).

    From https://tools.ietf.org/html/rfc5892#appendix-A.2:

      "This may occur in Indic scripts in a consonant-conjunct context
      (immediately following a virama), to control required display of
      such conjuncts."

    Args:
        value (str): String value to check.
        offset (int): Position within `value`.
        ucd (UnicodeData): Unicode character database.

    Returns:
        bool: True if value is allowed.
    """
    assert value[offset] == "\u200d"
    return ucd.combining_virama(_before(value, offset))


def rule_middle_dot(value, offset, ucd):
    """Return true if context permits a MIDDLE DOT (U+00B7).

    From https://tools.ietf.org/html/rfc5892#appendix-A.3:

      "Between 'l' (U+006C) characters only, used to permit the Catalan
      character ela geminada to be expressed."

    Args:
        value (str): String value to check.
        offset (int): Position within `value`.
        ucd (UnicodeData): Unicode character database.

    Returns:
        bool: True if value is allowed.
    """
    # pylint: disable=unused-argument
    assert value[offset] == "\u00b7"
    return 0x06C == _before(value, offset) == _after(value, offset)


def rule_greek_keraia(value, offset, ucd):
    """Return true if context permits GREEK LOWER NUMERAL SIGN (U+0375).

    From https://tools.ietf.org/html/rfc5892#appendix-A.4:

      "The script of the following character MUST be Greek."

    Args:
        value (str): String value to check.
        offset (int): Position within `value`.
        ucd (UnicodeData): Unicode character database.

    Returns:
        bool: True if value is allowed.
    """
    assert value[offset] == "\u0375"
    return ucd.greek_script(_after(value, offset))


def rule_hebrew_punctuation(value, offset, ucd):
    """Return true if context permits HEBREW PUNCTUATION GERESH or GERSHAYIM
    (U+05F3, U+05F4).

    From https://tools.ietf.org/html/rfc5892#appendix-A.5,
         https://tools.ietf.org/html/rfc5892#appendix-A.6:

      "The script of the preceding character MUST be Hebrew."

    Args:
        value (str): String value to check.
        offset (int): Position within `value`.
        ucd (UnicodeData): Unicode character database.

    Returns:
        bool: True if value is allowed.
    """
    assert value[offset] in "\u05f3\u05f4"
    return ucd.hebrew_script(_before(value, offset))


# These rules ignore the offset argument; they test the entire string. A string
# only needs to be tested once, irrespective of the number of times the rule is
# triggered.


def rule_katakana_middle_dot(value, offset, ucd):
    """Return true if context permits KATAKANA MIDDLE DOT (U+30FB).

    From https://tools.ietf.org/html/rfc5892#appendix-A.7:

      "Note that the Script of Katakana Middle Dot is not any of
      "Hiragana", "Katakana", or "Han".  The effect of this rule is to
      require at least one character in the label to be in one of those
      scripts."

    Args:
        value (str): String value to check.
        offset (int): Position within `value`.
        ucd (UnicodeData): Unicode character database.

    Returns:
        bool: True if value is allowed.
    """
    assert value[offset] == "\u30fb"
    return any(ucd.hiragana_katakana_han_script(ord(x)) for x in value)


def rule_arabic_indic(value, offset, ucd):
    """Return true if context permits ARABIC-INDIC DIGITS (U+0660..U+0669).

    From https://tools.ietf.org/html/rfc5892#appendix-A.8:

      "Can not be mixed with Extended Arabic-Indic Digits."

    Args:
        value (str): String value to check.
        offset (int): Position within `value`.
        ucd (UnicodeData): Unicode character database.

    Returns:
        bool: True if value is allowed.
    """
    assert ucd.arabic_indic(ord(value[offset]))
    return not any(ucd.extended_arabic_indic(ord(x)) for x in value)


def rule_extended_arabic_indic(value, offset, ucd):
    """Return true if context permits EXTENDED ARABIC-INDIC DIGITS
    (U+06F0..U+06F9).

    From https://tools.ietf.org/html/rfc5892#appendix-A.9:

      "Can not be mixed with Arabic-Indic Digits."

    Args:
        value (str): String value to check.
        offset (int): Position within `value`.
        ucd (UnicodeData): Unicode character database.

    Returns:
        bool: True if value is allowed.
    """
    assert ucd.extended_arabic_indic(ord(value[offset]))
    return not any(ucd.arabic_indic(ord(x)) for x in value)


_RULES = {
    0x200C: rule_zero_width_nonjoiner,
    0x200D: rule_zero_width_joiner,
    0x00B7: rule_middle_dot,
    0x0375: rule_greek_keraia,
    0x05F3: rule_hebrew_punctuation,
    0x05F4: rule_hebrew_punctuation,
    0x30FB: rule_katakana_middle_dot,
}


def _before(value, offset):
    """Return code point before `value[offset]` or raise IndexError."""
    if offset <= 0:
        raise IndexError(offset - 1)
    return ord(value[offset - 1])


def _after(value, offset):
    """Return code point after `value[offset]` or raise IndexError."""
    return ord(value[offset + 1])



"""Implements the PRECIS profile classes."""

import re

from precis_i18n.baseclass import FreeFormClass, IdentifierClass, raise_error
from precis_i18n.bidi import bidi_rule, has_rtl

# pylint: disable=no-self-use


class Profile:
    """Base class for a PRECIS profile.

    Subclasses should override the `*_rule` methods.

    Args:
        base (BaseClass): Base string class.
        name (str): Name of profile.
        casemap (Optional[str]): Case mapping function: 'fold' or 'lower'.
    """

    def __init__(self, base, name, casemap=None):
        self._base = base
        self._name = name
        # casemap can be either None, 'fold', or 'lower'.
        if casemap is None:
            self._casemap = None
        elif casemap == "fold":
            self._casemap = _casefold
        elif casemap == "lower":
            self._casemap = _caselower
        else:
            raise ValueError("Unknown casemap value: %s" % casemap)

    @property
    def base(self):
        """Base string class."""
        return self._base

    @property
    def name(self):
        """Profile name."""
        return self._name

    def enforce(self, value):
        """Ensure that all characters in `value` are allowed by the profile.

        If `value` is bytes, it's first decoded as UTF-8 to a string.

        Args:
            value (Union[str, bytes]): String value to enforce.

        Returns:
            str: Enforced value.

        Raises:
            UnicodeEncodeError: Value is disallowed by the profile.
            ValueError: `value` not a string or bytes.
        """
        # If we get called with a byte string, decode it first.
        if isinstance(value, bytes):
            value = value.decode("utf-8")
        elif not isinstance(value, str):
            raise ValueError("not a string")
        temp = self.apply_five_rules(value)
        temp = self.idempotence_check(temp)
        # Make sure the resulting value is not empty.
        if not temp:
            raise_error(self.name, value, -1, "empty")
        # Apply behavioral rules from the base string class last.
        return self.base.enforce(temp, self.name)

    def apply_five_rules(self, value):
        """Apply the five rules specified by RFC 8264 in order.

        Args:
            value (str): Value to enforce.

        Returns:
            str: Enforced value.
        """
        temp = self.width_mapping_rule(value)
        temp = self.additional_mapping_rule(temp)
        temp = self.case_mapping_rule(temp)
        temp = self.normalization_rule(temp)
        return self.directionality_rule(temp)

    def width_mapping_rule(self, value):
        """Apply width mapping rule.

        Args:
            value (str): Value to enforce.

        Returns:
            str: Enforced value.
        """
        return value

    def additional_mapping_rule(self, value):
        """Apply additional mapping rule.

        Args:
            value (str): Value to enforce.

        Returns:
            str: Enforced value.
        """
        return value

    def case_mapping_rule(self, value):
        """Apply case mapping rule.

        Args:
            value (str): Value to enforce.

        Returns:
            str: Enforced value.
        """
        if self._casemap:
            return self._casemap(value)
        return value

    def normalization_rule(self, value):
        """Apply normalization rule.

        Args:
            value (str): Value to enforce.

        Returns:
            str: Enforced value.
        """
        return self.base.ucd.normalize("NFC", value)

    def directionality_rule(self, value):
        """Apply directionality rule.

        Args:
            value (str): Value to enforce.

        Returns:
            str: Enforced value.
        """
        return value

    def idempotence_check(self, value):
        """Check that profile result is idempotent.

        Profiles that are not idempotent should override this method.

        Args:
            value (str): Value to enforce.

        Returns:
            str: Enforced value.
        """
        new_value = self.apply_five_rules(value)
        if new_value != value:
            raise_error(self.name, value, -1, "not_idempotent")
        return value


class Username(Profile):
    """Concrete class for Username profile.

    Reference:
        Name:  UsernameCasePreserved | UsernameCaseMapped

        Base Class:  IdentifierClass.

        Applicability:  Usernames in security and application protocols.

        Replaces:  The SASLprep profile of stringprep.

        Width-Mapping Rule:  Map fullwidth and halfwidth characters to their
           decomposition mappings.

        Additional Mapping Rule:  None.

        Case-Mapping Rule:  None | Map uppercase and titlecase characters to
           lowercase.

        Normalization Rule:  NFC.

        Directionality Rule:  The "Bidi Rule" defined in RFC 5893 applies.

        Enforcement:  To be defined by security or application protocols that
           use this profile.

        Specification:  RFC 8265, Section 3.3.

    Args:
        ucd (UnicodeData): Unicode character database.
        name (str): Name of profile.
        casemap (Optional[str]): Case mapping function: 'fold' or 'lower'.
    """

    def __init__(self, ucd, name, casemap=None):
        super().__init__(IdentifierClass(ucd), name, casemap)

    def width_mapping_rule(self, value):
        # Override
        return self.base.ucd.width_map(value)

    def directionality_rule(self, value):
        # Override
        # Only apply the "bidi rule" if the string contains RTL characters.
        if has_rtl(value, self.base.ucd):
            if not bidi_rule(value, self.base.ucd):
                raise_error(self.name, value, -1, "bidi_rule")
        return value


class OpaqueString(Profile):
    """Concrete class for OpaqueString profile.

    Reference:
        Name:  OpaqueString.

        Base Class:  FreeformClass.

        Applicability:  Passwords and other opaque strings in security and
           application protocols.

        Replaces:  The SASLprep profile of stringprep.

        Width-Mapping Rule:  None.

        Additional Mapping Rule:  Map non-ASCII space characters to ASCII
           space.

        Case-Mapping Rule:  None.

        Normalization Rule:  NFC.

        Directionality Rule:  None.

        Enforcement:  To be defined by security or application protocols that
           use this profile.

        Specification:  RFC 8265, Section 4.2.

    Args:
        ucd (UnicodeData): Unicode character database.
        name (str): Name of profile.
    """

    def __init__(self, ucd, name):
        super().__init__(FreeFormClass(ucd), name, casemap=None)

    def additional_mapping_rule(self, value):
        # Override
        return self.base.ucd.map_nonascii_space_to_ascii(value)


class Nickname(Profile):
    """Concrete class for Nickname profile.

    Reference:
        Name:  Nickname.

        Base Class:  FreeformClass.

        Applicability:  Nicknames in messaging and text conferencing
           technologies; petnames for devices, accounts, and people; and
           other uses of nicknames or petnames.

        Replaces:  None.

        Width Mapping Rule:  None (handled via NFKC).

        Additional Mapping Rule:  Map non-ASCII space characters to ASCII
           space, strip leading and trailing space characters, map interior
           sequences of multiple space characters to a single ASCII space.

        Case Mapping Rule:  Map uppercase and titlecase characters to
           lowercase using Unicode Default Case Folding.

        Normalization Rule:  NFKC.

        Directionality Rule:  None.

        Enforcement:  To be specified by applications.

        Specification:  RFC 8266

    Args:
        ucd (UnicodeData): Unicode character database.
        name (str): Name of profile.
        casemap (Optional[str]): Case mapping function: 'fold' or 'lower'.
    """

    def __init__(self, ucd, name, casemap=None):
        super().__init__(FreeFormClass(ucd), name, casemap)

    def additional_mapping_rule(self, value):
        # Override
        temp = self.base.ucd.map_nonascii_space_to_ascii(value)
        return re.sub(r"  +", " ", temp.strip(" "))

    def normalization_rule(self, value):
        # Override
        return self.base.ucd.normalize("NFKC", value)

    def idempotence_check(self, value):
        # Override
        # Nickname profile is not idempotent due to ordering of additional
        # and case mapping rules, so we apply them again.
        value = self.apply_five_rules(value)
        return super().idempotence_check(value)


def _casefold(s):
    return s.casefold()


def _caselower(s):
    return s.lower()



"""Implements the PRECIS string classes."""

from precis_i18n.context import context_rule_error
from precis_i18n.derived import CONTEXTJ, CONTEXTO, FREE_PVAL, PVALID, derived_property


class BaseClass:
    """Abstract base class for all String classes in PRECIS framework.

    Subclasses must set `_allowed` to a tuple of derived property names. For
    example, `_allowed = (PVALID, )`.

    Args:
        ucd (UnicodeData): Unicode character database.
        name (str): String class name.

    Attributes:
        ucd (UnicodeData): Unicode character database.
        name (str): String class name.
    """

    _allowed = ()

    def __init__(self, ucd, name=None):
        self.ucd = ucd
        self.name = name or self.__class__.__name__

    def enforce(self, value, codec_name=None):
        """Ensure that all characters in `value` are allowed by the string class.

        Args:
            value (str): String value to enforce.
            codec_name (Optional[str]): Codec name to report in exceptions. If
                None, use `self.name`.

        Returns:
            str: Enforced value.

        Raises:
            UnicodeEncodeError: Value is disallowed by the string class.
        """
        if codec_name is None:
            codec_name = self.name

        for i, char in enumerate(value):
            prop, kind = derived_property(ord(char), self.ucd)
            if prop in self._allowed:
                continue

            if prop in (CONTEXTJ, CONTEXTO):
                # Replace `kind` ('exceptions', 'join_control') with the
                # specific name of the context rule, if the rule fails.
                kind = context_rule_error(value, i, self.ucd)
                if not kind:
                    continue

            raise_error(codec_name, value, i, kind)

        return value


class IdentifierClass(BaseClass):
    """Concrete class representing PRECIS IdentifierClass from RFC 8264.

    Args:
        ucd (UnicodeData): Unicode character database.
        name (str): String class name.
    """

    _allowed = (PVALID,)


class FreeFormClass(BaseClass):
    """Concrete class repsenting PRECIS FreeFormClass from RFC 8264.

    Args:
        ucd (UnicodeData): Unicode character database.
        name (str): String class name.
    """

    _allowed = (PVALID, FREE_PVAL)


def raise_error(encoding, value, offset, error):
    """Raise specially formatted UnicodeEncodeError exception.

    Args:
        encoding (str): Name of the encoding/codec.
        value (str): Value being encoded.
        offset (int): Offset in `value` where error detected. Use -1 to
            indicate the entire string.
        error (str): Subtype of error detected.

    Raises:
        UnicodeEncodeError: Always.
    """
    if offset < 0:
        start = 0
        end = len(value)
    else:
        start = offset
        end = offset + 1

    reason = "DISALLOWED/%s" % error
    raise UnicodeEncodeError(encoding, value, start, end, reason)



from precis_i18n.unicode import UnicodeData

def context_rule_error(value: str, offset: int, ucd: UnicodeData) -> str: ...
def rule_zero_width_nonjoiner(value: str, offset: int, ucd: UnicodeData) -> bool: ...
def rule_zero_width_joiner(value: str, offset: int, ucd: UnicodeData) -> bool: ...
def rule_middle_dot(value: str, offset: int, ucd: UnicodeData) -> bool: ...
def rule_greek_keraia(value: str, offset: int, ucd: UnicodeData) -> bool: ...
def rule_hebrew_punctuation(value: str, offset: int, ucd: UnicodeData) -> bool: ...
def rule_katakana_middle_dot(value: str, offset: int, ucd: UnicodeData) -> bool: ...
def rule_arabic_indic(value: str, offset: int, ucd: UnicodeData) -> bool: ...
def rule_extended_arabic_indic(value: str, offset: int, ucd: UnicodeData) -> bool: ...

_RULES = ...



from typing import Set

from precis_i18n.unicode import UnicodeData

_LTR_FIRST = Set[str]
_LTR_ALLOWED = Set[str]
_LTR_LAST = Set[str]
_LTR_EXCL = Set[str]
_RTL_FIRST = Set[str]
_RTL_ALLOWED = Set[str]
_RTL_LAST = Set[str]
_RTL_EXCL = Set[str]
_RTL_ANY = Set[str]

def bidi_rule(value: str, ucd: UnicodeData) -> bool: ...
def has_rtl(value: str, ucd: UnicodeData) -> bool: ...
